/*
 *   JVending
 *   Copyright (C) 2004  Shane Isbell
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.jvending.provisioning.config;

import java.util.*;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

import java.util.logging.*;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.StringWriter;

import javax.provisioning.Capabilities;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;

import org.jvending.registry.Repository;
import org.jvending.registry.RepositoryRegistry;
import org.jvending.registry.jaxb.JaxbConfiguration;

import org.jvending.provisioning.config.devices.*;


/**
 * The <code>DeviceRepository</code> class is responsible for looking up the
 * appropriate capabilities and adapters for a device.
 *
 * @author Shane Isbell
 * @since 1.2a
 */
public final class DeviceRepository implements Repository {

    private static Logger logger = Logger.getLogger("DeviceRepository");//last log 19

    private List deviceMappings;

    private List devices;

    private Devices devicesCopy;

    public void setRepositoryRegistry(RepositoryRegistry repositoryRegistry) {
        //don't need this
    }    

    public void load(InputStream inputStream, Hashtable properties) throws IOException {
        if (properties == null) throw new IOException("JV-1303-001: Null properties values for the device repository");
        if (inputStream == null) throw new IOException("JV-1303-002: Null inputstream for the device repository");

        String _package = (String) properties.get("binding-package");
        byte[] inputBytes = copyStreamToBytes(inputStream);

        Devices device = (Devices) JaxbConfiguration.parse(new ByteArrayInputStream(copyBytes(inputBytes)), _package);
        devicesCopy = (Devices) JaxbConfiguration.parse(new ByteArrayInputStream(copyBytes(inputBytes)), _package);

        List deviceMapping = device.getDeviceMapping();
        devices = Collections.unmodifiableList(device.getDevice());

        DeviceMappingMapper deviceMappingMapper = new DeviceMappingMapper();
        deviceMappings = Collections.unmodifiableList(deviceMappingMapper.map(deviceMapping));
    }

    /**
     * Matches the device id based upon the HTTP headers.
     */
    public String getDeviceId(Map requestMap) {
        for (Iterator i = deviceMappings.iterator(); i.hasNext();) {
            DeviceMapping mapping = (DeviceMapping) i.next();
            DeviceMatch match = mapping.getDeviceMatch();
            int j = match.match(requestMap);
            if (j == 1) {
                String id = mapping.getIdentifier();
                return (id != null) ? id.trim() : id;
            }
        }
        return null;
    }

    public Set getDeviceIds() {
        Set set = new HashSet();
        for (Iterator i = deviceMappings.iterator(); i.hasNext();) {
            DeviceMapping mapping = (DeviceMapping) i.next();
            String id = mapping.getIdentifier();
            if (id != null) set.add(id.trim());
        }
        return Collections.unmodifiableSet(set);//safe
    }

    /**
     * Accessor capabilities of the device that matches the HTTP headers.
     *
     * @param requestMap <code>Map</code> containing the HTTP request headers
     * @return <code>Capabilities</code> of the device that matched the requestMap
     */
    public Capabilities getCapabilities(Map requestMap) {
        String id = getDeviceId(requestMap);
        logger.finest("JV-1303-003: DEVICE ID:: " + id);
        if (id == null) {
            logger.finest("JV-1303-004: Http Headers = " + requestMap);
            return null;
        }

        return getCapabilitiesFor(id.trim());
    }

    /**
     * Accessor for the capabilities of the device.
     *
     * @param deviceId the identifier of the device
     * @return Capabilities of the device for the given id
     */
    public Capabilities getCapabilitiesFor(String deviceId) {
        Map capabilityMap = getCapabilitiesForPrivate(deviceId);
        return (capabilityMap != null) ? new CapabilitiesImpl(capabilityMap) : null;
    }

    /**
     * Gets the names of the adapters that can handle the provisioning interaction for the device. If deviceId is
     * null or empty, returns all adapters names. If deviceId is not found, returns an empty list.
     *
     * @param deviceId the identifier of the device
     * @return <code>List</code> that contains <code>String</code> objects denoting the names of the adapters.
     */
    public List getAdapterNamesFor(String deviceId) {
        if (deviceId == null || deviceId.trim().equals("")) {
            logger.finest("JV-1303-005: Device Id is Null. Returning empty list of adapter names.");
            return Collections.EMPTY_LIST;
        }
        //TO DO: Check this method for any null pointer exceptions
        for (Iterator i = devices.iterator(); i.hasNext();) {
            DeviceType deviceType = (DeviceType) i.next();
            List id = deviceType.getIdentifier();

            for (Iterator j = id.iterator(); j.hasNext();) {
                String identifier = ((String) j.next()).trim();
                if (identifier.equals(deviceId.trim())) {
                    List adapterNames = new ArrayList();

                    for (Iterator k = deviceType.getAdapterName().iterator(); k.hasNext();) {
                        String adapterName = (String) k.next();
                        adapterNames.add(adapterName.trim());
                    }
                    return Collections.unmodifiableList(adapterNames);
                }
            }

        }
        logger.finest("JV-1303-006: Device Id not found: Device Id = " + deviceId.trim());
        return Collections.EMPTY_LIST;
    }


    String getDeviceMapping(String deviceId) {
        StringWriter writer = new StringWriter();

        for (Iterator i = deviceMappings.iterator(); i.hasNext();) {
            DeviceMapping mapping = (DeviceMapping) i.next();
            String id = mapping.getIdentifier();
            if (id != null && id.equals(deviceId)) {
                try {
                    JAXBContext jc = JAXBContext.newInstance("org.jvending.provisioning.config.devices");
                    Marshaller m = jc.createMarshaller();
                    m.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
                    m.marshal(mapping, writer);
                    return writer.toString();
                } catch (JAXBException e) {
                    logger.info("JV-1303-007: Could not marshal device mapping: Device Id = " + deviceId);
                }
            }
        }
        return null;
    }


    public DeviceConfig getDeviceConfig() {
        return new DeviceConfigImpl();
    }

    private byte[] copyStreamToBytes(InputStream is) throws IOException {
        ByteArrayOutputStream os = new ByteArrayOutputStream();
        byte[] buffer = new byte[1024];
        int n = 0;
        while ((n = is.read(buffer)) >= 0) {
            os.write(buffer, 0, n);
        }
        return os.toByteArray();
    }

    private byte[] copyBytes(byte[] input) {
        byte[] output = new byte[input.length];
        System.arraycopy(input, 0, output, 0, output.length);
        return output;
    }

    //returns copy
    private Map getCapabilitiesForPrivate(String deviceId) {
        if (deviceId == null) return null;

        for (Iterator i = devices.iterator(); i.hasNext();) {
            DeviceType deviceType = (DeviceType) i.next();
            List id = deviceType.getIdentifier();

            for (Iterator j = id.iterator(); j.hasNext();) {
                String identifier = ((String) j.next());
                if (identifier != null && identifier.trim().equals(deviceId.trim())) {
                    return toMap(deviceType.getCapability());//TO DO: optimize
                }
            }
        }
        return null;
    }

    private Map toMap(List capabilities) {
        Map map = new HashMap();
        for (Iterator i = capabilities.iterator(); i.hasNext();) {
            CapabilityType capabilityType = (CapabilityType) i.next();
            String name = capabilityType.getCapabilityName();
            name = (name != null) ? name.trim() : "";

            String value = capabilityType.getCapabilityValue();
            value = (value != null) ? value.trim() : "";

            String[] tokenValues = value.split("[,]");
            int size = tokenValues.length;
            List values = new ArrayList();
            for (int j = 0; j < size; j++) {
                values.add(tokenValues[j].trim());
            }
            map.put(name, values);
        }
        return Collections.unmodifiableMap(map);
    }

    private static class DeviceMappingMapper {
        List map(List deviceMappings) throws IOException {
            List dMap = new ArrayList();

            for (Iterator i = deviceMappings.iterator(); i.hasNext();) {
                DeviceMapping deviceMapping = null;
                DeviceMappingType deviceMappingType = (DeviceMappingType) i.next();
                String id = deviceMappingType.getIdentifier();

                RequestMappingType requestMappingType = deviceMappingType.getRequestMapping();
                if (requestMappingType != null) {
                    deviceMapping = new DeviceMapping(id, new MatchTerminal(requestMappingType));
                    dMap.add(deviceMapping);
                    continue;
                }

                MatchAllType matchAllType = deviceMappingType.getMatchAll();
                if (matchAllType != null) {
                    deviceMapping = new DeviceMapping(id, new MatchAll(matchAllType));
                    dMap.add(deviceMapping);
                    continue;
                }

                MatchAnyType matchAnyType = deviceMappingType.getMatchAny();
                if (matchAnyType != null) {
                    deviceMapping = new DeviceMapping(id, new MatchAny(matchAnyType));
                    dMap.add(deviceMapping);
                    continue;
                }

                MatchNotType matchNotType = deviceMappingType.getMatchNot();
                if (matchNotType != null) {
                    deviceMapping = new DeviceMapping(id, new MatchNot(matchNotType));
                    dMap.add(deviceMapping);
                    continue;
                }
            }
            return dMap;
        }
    }

    private static class MatchAll extends DeviceMatch {

        List terminals;

        List nots;

        List anys;

        MatchAll(MatchAllType all) {
            List list = all.getRequestMappingOrMatchAnyOrMatchNot();

            terminals = new ArrayList();
            nots = new ArrayList();
            anys = new ArrayList();

            for (Iterator i = list.iterator(); i.hasNext();) {
                Object o = i.next();
                if (o instanceof MatchNotType)
                    nots.add(o);
                if (o instanceof MatchAnyType)
                    anys.add(o);
                if (o instanceof RequestMappingType)
                    terminals.add(o);
            }
        }

        /**
         * Matches the request to a match-all rule.
         *
         * @param requestMap <code>Map</code> containing request headers.
         * @return 0 denotes no match, 1 denotes a match
         */
        int match(Map requestMap) {
            int result = matchAllTerminals(terminals, requestMap);
            if (result == 0) return 0;

            result = matchAny(anys, requestMap);
            if (result == 0) return 0;

            result = matchNot(nots, requestMap);
            if (result == 0) return 0;

            return 1;
        }
    }

    private static class MatchAny extends DeviceMatch {

        List terminals;

        List nots;

        List alls;

        MatchAny(MatchAnyType any) {
            List list = any.getRequestMappingOrMatchAllOrMatchNot();

            terminals = new ArrayList();
            nots = new ArrayList();
            alls = new ArrayList();

            for (Iterator i = list.iterator(); i.hasNext();) {
                Object o = i.next();
                if (o instanceof MatchNotType)
                    nots.add(o);
                if (o instanceof MatchAllType)
                    alls.add(o);
                if (o instanceof RequestMappingType)
                    terminals.add(o);
            }
        }

        /**
         * Matches the request to a match-any rule.
         *
         * @param requestMap <code>Map</code> containing request headers.
         * @return 0 denotes no match, 1 denotes a match
         */
        int match(Map requestMap) {
            int result = matchAnyTerminal(terminals, requestMap);
            if (result == 1) return 1;
            if (alls.size() > 0) {
                result = matchAll(alls, requestMap);
                if (result == 1) return 1;
            }

            if (nots.size() > 0) {
                result = matchNot(nots, requestMap);
                if (result == 1) return 1;
            }
            return 0;
        }
    }

    private static class MatchNot extends DeviceMatch {

        RequestMappingType requestMapping;

        MatchAllType all;

        MatchAnyType any;

        MatchNot(MatchNotType not) {
            this.requestMapping = not.getRequestMapping();
            this.all = not.getMatchAll();
            this.any = not.getMatchAny();
        }

        /**
         * Matches the request to a match-not rule.
         *
         * @param requestMap <code>Map</code> containing request headers.
         * @return 0 denotes no match, 1 denotes a match
         */
        int match(Map requestMap) {
            int result = matchTerminal(requestMapping, requestMap);
            if (result == 1) return 0;

            List alls = new ArrayList();
            alls.add(all);
            result = matchAll(alls, requestMap);
            if (result == 1) return 0;

            List anys = new ArrayList();
            anys.add(any);
            result = matchAny(anys, requestMap);
            if (result == 1) return 0;
            return 1;
        }
    }

    private static class MatchTerminal extends DeviceMatch {

        private String terminalName;

        private Pattern pattern;

        MatchTerminal(RequestMappingType terminal) {
            pattern = Pattern.compile(terminal.getHeaderValue().trim());
            terminalName = terminal.getHeaderName().trim();
        }

        int match(Map requestMap) {
            //Will not match any device, in this case
            if (requestMap == null || !requestMap.containsKey(terminalName)) {
                logger.finest("JV-1303-008: Terminal Name = " + terminalName + ", Map = " + requestMap);
                return 0;
            }

            String headerValue = (String) requestMap.get(terminalName);
            if (headerValue == null) return 0;
            headerValue = headerValue.trim();
            logger.finest("JV-1303-009: Terminal Name = " + terminalName + ", Header Value = " + headerValue);
            Matcher matcher = pattern.matcher(headerValue);
            return (matcher.matches()) ? 1 : 0;
        }
    }

    private static abstract class DeviceMatch {

        abstract int match(Map requestMap);

        int matchTerminal(RequestMappingType terminal, Map requestMap) {
            MatchTerminal mt = new MatchTerminal(terminal);
            int x = mt.match(requestMap);
            logger.finest("JV-1303-010: Match terminal: Value = " + x);
            return x;
            // return mt.match(requestMap);
        }

        int matchAnyTerminal(List terminals, Map requestMap) {
            int x = 0;
            for (Iterator i = terminals.iterator(); i.hasNext();) {
                x = matchTerminal((RequestMappingType) i.next(), requestMap);
                logger.finest("JV-1303-011: Match any terminal: Value = " + x);
                if (x > 0) return 1;
            }
            return 0;
        }

        int matchAllTerminals(List terminals, Map requestMap) {
            int x = 0;
            for (Iterator i = terminals.iterator(); i.hasNext();) {
                x = x + matchTerminal((RequestMappingType) i.next(), requestMap);
            }
            return (x == terminals.size()) ? 1 : 0;
        }

        int matchNot(List nots, Map requestMap) {
            int x = 0;

            for (Iterator t = nots.iterator(); t.hasNext();) {
                MatchNot ma = new MatchNot((MatchNotType) t.next());
                x = x + ma.match(requestMap);
            }
            return (x == nots.size()) ? 1 : 0;
        }

        int matchAll(List alls, Map requestMap) {
            int x = 0;

            for (Iterator t = alls.iterator(); t.hasNext();) {
                MatchAll ma = new MatchAll((MatchAllType) t.next());
                x = x + ma.match(requestMap);
            }
            return (x == alls.size()) ? 1 : 0;
        }

        int matchAny(List anys, Map requestMap) {
            int x = 0;

            for (Iterator t = anys.iterator(); t.hasNext();) {
                MatchAny ma = new MatchAny((MatchAnyType) t.next());
                x = x + ma.match(requestMap);
                logger.finest("JV-1303-012: Match Any: Value = " + x);
                if (x > 0) return 1;
            }
            return 0;
        }
    }

    private static class DeviceMapping {

        private String id;

        private DeviceMatch deviceMatch;

        DeviceMapping(String id, DeviceMatch deviceMatch) {
            this.id = id;
            this.deviceMatch = deviceMatch;
        }

        String getIdentifier() {
            return id;
        }

        DeviceMatch getDeviceMatch() {
            return deviceMatch;
        }
    }

    private static class CapabilitiesImpl implements Capabilities {

        private final Map capability;

        //map of capabilites for a specific device
        public CapabilitiesImpl(Map capability) {
            this.capability = new HashMap(capability);
        }

        //List of Strings: Values of capability
        public List getCapability(String name) {
            if (name == null) return null;
            List list = (List) capability.get(name);
            return (list == null) ? null : Collections.unmodifiableList(list);
        }

        public Set getCapabilityNames() {
            Set c = capability.keySet();
            if (c == null) c = new TreeSet();
            return Collections.unmodifiableSet(new TreeSet(c));
        }

    }

    private class DeviceConfigImpl implements DeviceConfig {

        private ObjectFactory factory = new ObjectFactory();

        public DevicesType getDevices() {
            return devicesCopy;
        }

        public void addDevice(String deviceId, String[] adapters, Map capabilities, String userAgent) throws IOException {
            try {
                DeviceType device = createDevice(deviceId, adapters, capabilities);
                DeviceMappingType deviceMapping = createDeviceMapping(deviceId, userAgent);
                if (device != null && deviceMapping != null) {
                    devicesCopy.getDeviceMapping().add(deviceMapping);
                    devicesCopy.getDevice().add(device);
                }
            } catch (JAXBException e) {
                logger.config("JV-1303-013: Failed to add device: Device Id = " + deviceId);
                throw new IOException();
            }
        }

        public boolean deleteDevice(String deviceId) {
            return deleteDeviceInternal(deviceId);
        }

        private DeviceType createDevice(String deviceId, String[] adapters, Map capabilities) throws JAXBException {
            DeviceType device = factory.createDeviceType();
            device.getIdentifier().add(deviceId);
            int adapterSize = adapters.length;
            for (int i = 0; i < adapterSize; i++) {
                device.getAdapterName().add(adapters[i]);
            }
            return device;
        }

        private DeviceMappingType createDeviceMapping(String deviceId, String userAgent) throws JAXBException {
            DeviceMappingType deviceMapping = factory.createDeviceMappingType();
            RequestMappingType requestMapping = factory.createRequestMappingType();
            requestMapping.setHeaderName("user-agent");
            requestMapping.setHeaderValue(userAgent);

            deviceMapping.setIdentifier(deviceId);
            deviceMapping.setRequestMapping(requestMapping);
            return deviceMapping;
        }

        private boolean deleteDeviceMapping(String deviceId) {
            if (deviceId == null) {
                logger.info("JV-1303-014: Device id is null. Nothing to delete");
                return false;
            }
            List deviceMappings = devicesCopy.getDeviceMapping();

            for (Iterator i = deviceMappings.iterator(); i.hasNext();) {
                DeviceMappingType deviceMappingType = (DeviceMappingType) i.next();
                String id = deviceMappingType.getIdentifier();
                if (id != null && id.trim().equals(deviceId)) {
                    i.remove();
                    logger.config("JV-1303-015: Removed Device Mapping: Device Id = " + deviceId);
                    return true;
                }
            }
            logger.info("JV-1303-016: Could not find device mapping: Device Id = " + deviceId);
            return false;
        }

        private boolean deleteDeviceInternal(String deviceId) {
            if (deviceId == null) {
                logger.info("JV-1303-017: Device id is null. Nothing to delete");
                return false;
            }
            List devices = devicesCopy.getDevice();

            for (Iterator i = devices.iterator(); i.hasNext();) {
                DeviceType deviceType = (DeviceType) i.next();
                List ids = deviceType.getIdentifier();

                for (Iterator j = ids.iterator(); j.hasNext();) {
                    String id = (String) j.next();
                    if (id != null && id.trim().equals(deviceId)) {
                        if (deleteDeviceMapping(deviceId)) {
                            i.remove();
                            logger.config("JV-1303-018: Removed Device: Device Id = " + deviceId);
                            return true;
                        } else {
                            return false;
                        }
                    }
                }
            }
            logger.info("JV-1303-019: Could not find device: Device Id = " + deviceId);
            return false;
        }
    }
}