/*
 *   JVending
 *   Copyright (C) 2005  Shane Isbell
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.jvending.provisioning.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.*;

import java.lang.reflect.Constructor;

import javax.provisioning.AdapterInfo;

import javax.provisioning.adapter.Adapter;
import javax.provisioning.adapter.AdapterConfig;
import javax.provisioning.adapter.AdapterContext;

import org.jvending.provisioning.config.adapters.*;
import org.jvending.provisioning.config.AdapterRepository;

/**
 * @author Shane Isbell
 * @since 2.0.0
 */
public final class AdapterRepositoryWrapper {//TODO: Fix: Exposed as public for the stocking tool, create mediator
//load adaptercontext here

    private static Logger logger = Logger.getLogger("AdapterRepositoryWrapper");

    private Map adapterConfigs;

    private List adapterInfos;

    private Map adapters;

    private List adapterNames;

    private AdapterContext adapterContext;

    public AdapterRepositoryWrapper(AdapterRepository adapterRepository, AdapterContext adapterContext) {
        if (adapterRepository == null) {
            logger.severe("JV-1503-001: The AdapterRepository is null");
            return;
        }

        if (adapterContext == null) {
            logger.severe("JV-1503-002: The AdapterContext is null");
            return;
        }

        this.adapterContext = adapterContext;
        adapterConfigs = new HashMap();
        adapterInfos = new ArrayList();
        adapters = new HashMap();
        adapterNames = new ArrayList();

        List adapterTypes = adapterRepository.getAdapters();
        for (Iterator i = adapterTypes.iterator(); i.hasNext();) {
            AdapterType adapterType = (AdapterType) i.next();
            String adapterName = adapterType.getAdapterName().trim();
            adapterNames.add(adapterName);
            long duration = 0L;
            try {
                duration = Long.parseLong(adapterType.getFulfillmentDuration().trim());
            } catch (NumberFormatException e) {
                logger.info("JV-1503-003: Problem with fulfillment duration. Setting to never expire.");
            }

            DescriptorFileType df = adapterType.getDescriptorFile();
            String fileExtension = (df != null) ? df.getExtension().trim() : null;
            String mimeType = (df != null) ? df.getMimeType().trim() : null;
            String baseUri = adapterType.getBaseUri().trim();

            Map initParam = toMap(adapterType.getInitParam());
            String className = adapterType.getAdapterClass().trim();

            AdapterConfig adapterConfig = AdapterFactory.createAdapterConfig(adapterName,
                    duration, fileExtension, mimeType, baseUri, initParam, className,
                    adapterContext
            );

            adapterConfigs.put(adapterName, adapterConfig);//TODO: Store only once
            adapterInfos.add((AdapterInfo) adapterConfig);
            //Spec: There must be only one instance of an adapter per JVM. This impl ok if the wrapper is only
            //invoked once. This impl makes it difficult, however, to destroy and adapter and then re init it.
            adapters.put(adapterName, getAdapterFor(className, adapterConfig));//could put in null adapter here.
            logger.info("JV-1503-004: Loaded adapter: ClassName = " + className + ", " + adapterConfig.toString());
        }
    }

    public List getAdapterNames() {
        return Collections.unmodifiableList(adapterNames);
    }

    public AdapterConfig getAdapterConfigFor(String adapterName) {
        return (AdapterConfig) adapterConfigs.get(adapterName);
    }

    public AdapterInfo getAdapterInfoFor(String adapterName) {
        if (adapterName == null) return null;
        else adapterName = adapterName.trim();

        for (Iterator i = adapterInfos.iterator(); i.hasNext();) {
            AdapterInfo adapterInfo = (AdapterInfo) i.next();
            String adapterName2 = adapterInfo.getAdapterName();
            if (adapterName.equals(adapterName2)) return adapterInfo;
        }
        return null;
    }

    public Adapter getAdapterFor(String adapterName) {
        return (Adapter) adapters.get(adapterName);//could return null
    }

    public List getAdapterInfos() {
        return Collections.unmodifiableList(adapterInfos);
    }

    public AdapterContext getAdapterContext() {
        return adapterContext;
    }


    private Map toMap(List list) {
        Map map = new HashMap();
        for (Iterator i = list.iterator(); i.hasNext();) {
            InitParamType type = (InitParamType) i.next();
            String name = type.getParamName();
            String value = type.getParamValue();
            map.put(name.trim(), value.trim());
        }
        return map;
    }

    private Adapter getAdapterFor(String className, AdapterConfig adapterConfig) {
        Adapter adapter = null;
        try {
            Class adapterClass = Class.forName(className);
            Constructor adapterConstructor = adapterClass.getConstructor(null);
            adapter = (Adapter) adapterConstructor.newInstance(null);
            adapter.init(adapterConfig);//Spec, to initialize
        } catch (Exception e) {
            e.printStackTrace();
            logger.info("JV-1503-005: Could not find or intialize adapter: " + className);
        }
        return adapter;

    }


}